"""Collection of different help functions."""
__author__ = 'Lukas Leufen, Felix Kleinert'
__date__ = '2019-10-21'

import inspect
import math
import argparse

import numpy as np
import xarray as xr
import dask.array as da

from typing import Dict, Callable, Union, List, Any, Tuple

from tensorflow.keras.models import Model
from tensorflow.python.keras.layers import deserialize, serialize
from tensorflow.python.keras.saving import saving_utils

"""
The following code is copied from: https://github.com/tensorflow/tensorflow/issues/34697#issuecomment-627193883
and is a hotfix to make keras.model.model models serializable/pickable
"""


def unpack(model, training_config, weights):
    restored_model = deserialize(model)
    if training_config is not None:
        restored_model.compile(
            **saving_utils.compile_args_from_training_config(
                training_config
            )
        )
    restored_model.set_weights(weights)
    return restored_model

# Hotfix function
def make_keras_pickable():

    def __reduce__(self):
        model_metadata = saving_utils.model_metadata(self)
        training_config = model_metadata.get("training_config", None)
        model = serialize(self)
        weights = self.get_weights()
        return (unpack, (model, training_config, weights))

    cls = Model
    cls.__reduce__ = __reduce__


" end of hotfix "


def to_list(obj: Any) -> List:
    """
    Transform given object to list if obj is not already a list. Sets are also transformed to a list.

    :param obj: object to transform to list

    :return: list containing obj, or obj itself (if obj was already a list)
    """
    if isinstance(obj, (set, tuple, type({}.keys()))):
        obj = list(obj)
    elif not isinstance(obj, list):
        obj = [obj]
    return obj


def sort_like(list_obj: list, sorted_obj: list):
    """
    Sort elements of list_obj as ordered in sorted_obj. Length of sorted_obj as allowed to be higher than length of
    list_obj, but must contain at least all objects of list_obj. Will raise AssertionError, if not all elements of
    list_obj are also in sorted_obj. Also it is required for list_obj and sorted_obj to have only unique elements.

    :param list_obj: list to sort
    :param sorted_obj: list to use ordering from

    :return: sorted list
    """
    assert set(list_obj).issubset(sorted_obj)
    assert len(set(list_obj)) == len(list_obj)
    assert len(set(sorted_obj)) == len(sorted_obj)
    return [e for e in sorted_obj if e in list_obj]


def dict_to_xarray(d: Dict, coordinate_name: str) -> xr.DataArray:
    """
    Convert a dictionary of 2D-xarrays to single 3D-xarray. The name of new coordinate axis follows <coordinate_name>.

    :param d: dictionary with 2D-xarrays
    :param coordinate_name: name of the new created axis (2D -> 3D)

    :return: combined xarray
    """
    if len(d.keys()) == 1:
        k = list(d.keys())
        xarray: xr.DataArray = d[k[0]]
        return xarray.expand_dims(dim={coordinate_name: k}, axis=0)
    else:
        xarray = None
        for k, v in d.items():
            if xarray is None:
                xarray = v
                xarray.coords[coordinate_name] = k
            else:
                tmp_xarray = v
                tmp_xarray.coords[coordinate_name] = k
                xarray = xr.concat([xarray, tmp_xarray], coordinate_name)
        return xarray


def float_round(number: float, decimals: int = 0, round_type: Callable = math.ceil) -> float:
    """
    Perform given rounding operation on number with the precision of decimals.

    :param number: the number to round
    :param decimals: numbers of decimals of the rounding operations (default 0 -> round to next integer value)
    :param round_type: the actual rounding operation. Can be any callable function like math.ceil, math.floor or python
        built-in round operation.

    :return: rounded number with desired precision
    """
    multiplier = 10. ** decimals
    return round_type(number * multiplier) / multiplier


def relative_round(x: float, sig: int, ceil=False, floor=False) -> float:
    """
    Round small numbers according to given "significance".

    Example: relative_round(0.03112, 2) -> 0.031, relative_round(0.03112, 1) -> 0.03

    :params x: number to round
    :params sig: "significance" to determine number of decimals

    :return: rounded number
    """
    assert sig >= 1
    assert not (ceil and floor)
    if x == 0:
        return 0
    else:
        rounded_number = round(x, sig-get_order(x)-1)
        if floor is True and rounded_number > round(x, sig-get_order(x)):
            res = rounded_number - 10 ** (get_order(x) - sig + 1)
        elif ceil is True and rounded_number < round(x, sig-get_order(x)):
            res = rounded_number + 10 ** (get_order(x) - sig + 1)
        else:
            res = rounded_number
        return round(res, sig-get_order(res)-1)


def get_order(x: float):
    """Get order of number (as power of 10)"""
    if x == 0:
        return -np.inf
    else:
        return int(np.floor(np.log10(abs(x))))


def remove_items(obj: Union[List, Dict, Tuple], items: Any):
    """
    Remove item(s) from either list, tuple or dictionary.

    :param obj: object to remove items from (either dictionary or list)
    :param items: elements to remove from obj. Can either be a list or single entry / key

    :return: object without items
    """

    def remove_from_list(list_obj, item_list):
        """Remove implementation for lists."""
        if len(item_list) > 1:
            return [e for e in list_obj if e not in item_list]
        elif len(item_list) == 0:
            return list_obj
        else:
            list_obj = list_obj.copy()
            try:
                list_obj.remove(item_list[0])
            except ValueError:
                pass
            return list_obj

    def remove_from_dict(dict_obj, key_list):
        """Remove implementation for dictionaries."""
        return {k: v for k, v in dict_obj.items() if k not in key_list}

    items = to_list(items)
    if isinstance(obj, list):
        return remove_from_list(obj, items)
    elif isinstance(obj, dict):
        return remove_from_dict(obj, items)
    elif isinstance(obj, tuple):
        return tuple(remove_from_list(to_list(obj), items))
    else:
        raise TypeError(f"{inspect.stack()[0][3]} does not support type {type(obj)}.")


def select_from_dict(dict_obj: dict, sel_list: Any, remove_none: bool = False, filter_cond: bool = True) -> dict:
    """
    Extract all key values pairs whose key is contained in the sel_list.

    Does not perform a check if all elements of sel_list are keys of dict_obj. Therefore, the number of pairs in the
    returned dict is always smaller or equal to the number of elements in the sel_list. If `filter_cond` is given, this
    method either return the parts of the input dictionary that are included or not in `sel_list`.
    """
    sel_list = to_list(sel_list)
    assert isinstance(dict_obj, dict)
    sel_dict = {k: v for k, v in dict_obj.items() if (k in sel_list) is filter_cond}
    sel_dict = sel_dict if not remove_none else {k: v for k, v in sel_dict.items() if v is not None}
    return sel_dict


def extract_value(encapsulated_value):
    try:
        if isinstance(encapsulated_value, str):
            raise TypeError
        if len(encapsulated_value) == 1:
            return extract_value(encapsulated_value[0])
        else:
            raise NotImplementedError("Trying to extract an encapsulated value from objects with more than a single "
                                      "entry is not supported by this function.")
    except TypeError:
        return encapsulated_value


def is_xarray(arr) -> bool:
    """
    Returns True if arr is xarray.DataArray or xarray.Dataset.
    :param arr: variable in question
    :type arr: Any
    :return:
    :rtype: bool
    """
    return isinstance(arr, xr.DataArray) or isinstance(arr, xr.Dataset)


def convert2xrda(arr: Union[xr.DataArray, xr.Dataset, np.ndarray, int, float],
                 use_1d_default: bool = False, **kwargs) -> Union[xr.DataArray, xr.Dataset]:
    """
    Converts np.array, int or float object to xarray.DataArray.

    If a xarray.DataArray or xarray.Dataset is passed, returns that unchanged.
    :param arr:
    :type arr: xr.DataArray, xr.Dataset, np.ndarray, int, float
    :param use_1d_default:
    :type use_1d_default: bool
    :param kwargs: Any additional kwargs which are accepted by xr.DataArray()
    :type kwargs:
    :return:
    :rtype: xr.DataArray, xr.DataSet
    """
    if is_xarray(arr):
        return arr
    else:
        if use_1d_default:
            if isinstance(arr, da.core.Array):
                raise TypeError(f"`use_1d_default=True' is used with `arr' of type da.array. For da.arrays please "
                                f"pass `use_1d_default=False' and specify keywords for xr.DataArray via kwargs.")
            dims = kwargs.pop('dims', 'points')
            coords = kwargs.pop('coords', None)
            try:
                if coords is None:
                    coords = {dims: range(arr.shape[0])}
            except (AttributeError, IndexError):
                if isinstance(arr, int) or isinstance(arr, float):
                    coords = kwargs.pop('coords', {dims: range(1)})
                    dims = to_list(dims)
                else:
                    raise TypeError(f"`arr' must be arry-like, int or float. But is of type {type(arr)}")
            kwargs.update({'dims': dims, 'coords': coords})

        return xr.DataArray(arr, **kwargs)


def filter_dict_by_value(dictionary: dict, filter_val: Any, filter_cond: bool) -> dict:
    """
    Filter dictionary by its values.

    :param dictionary: dict to filter
    :param filter_val: search only for key value pair with a value equal to filter_val
    :param filter_cond: indicate to use either all dict entries that fulfil the filter_val criteria (if `True`) or that
        do not match the criteria (if `False`)
    :returns: a filtered dict with either matching or non-matching elements depending on the `filter_cond`
    """
    filter_val = to_list(filter_val)
    return dict(filter(lambda x: (x[1] in filter_val) is filter_cond, dictionary.items()))


def str2bool(v):
    if isinstance(v, bool):
        return v
    elif isinstance(v, str):
        if v.lower() in ('yes', 'true', 't', 'y', '1'):
            return True
        elif v.lower() in ('no', 'false', 'f', 'n', '0'):
            return False
        else:
            raise argparse.ArgumentTypeError('Boolean value expected.')
    else:
        raise argparse.ArgumentTypeError('Boolean value expected.')


def squeeze_coords(d):
    """Look for unused coords and remove them. Does only work for xarray DataArrays."""
    try:
        return d.drop(set(d.coords.keys()).difference(d.dims))
    except Exception:
        return d


# def convert_size(size_bytes):
#     if size_bytes == 0:
#         return "0B"
#     size_name = ("B", "KB", "MB", "GB", "TB", "PB", "EB", "ZB", "YB")
#     i = int(math.floor(math.log(size_bytes, 1024)))
#     p = math.pow(1024, i)
#     s = round(size_bytes / p, 2)
#     return "%s %s" % (s, size_name[i])
#
#
# def get_size(obj, seen=None):
#     """Recursively finds size of objects"""
#     size = sys.getsizeof(obj)
#     if seen is None:
#         seen = set()
#     obj_id = id(obj)
#     if obj_id in seen:
#         return 0
#     # Important mark as seen *before* entering recursion to gracefully handle
#     # self-referential objects
#     seen.add(obj_id)
#     if isinstance(obj, dict):
#         size += sum([get_size(v, seen) for v in obj.values()])
#         size += sum([get_size(k, seen) for k in obj.keys()])
#     elif hasattr(obj, '__dict__'):
#         size += get_size(obj.__dict__, seen)
#     elif hasattr(obj, '__iter__') and not isinstance(obj, (str, bytes, bytearray)):
#         size += sum([get_size(i, seen) for i in obj])
#     return size
