__author__ = "Lukas Leufen"
__date__ = '2021-02-'

from functools import reduce, partial
from typing import Union

from mlair.model_modules import AbstractModelClass
from mlair.helpers import select_from_dict, to_list
from mlair.model_modules.loss import var_loss, custom_loss
from mlair.model_modules.advanced_paddings import PadUtils, Padding2D, SymmetricPadding2D

import tensorflow.keras as keras


class CNNfromConfig(AbstractModelClass):
    _activation = {"relu": keras.layers.ReLU, "tanh": partial(keras.layers.Activation, "tanh"),
                   "sigmoid": partial(keras.layers.Activation, "sigmoid"),
                   "linear": partial(keras.layers.Activation, "linear"),
                   "prelu": partial(keras.layers.PReLU, alpha_initializer=keras.initializers.constant(value=0.25)),
                   "leakyrelu": keras.layers.LeakyReLU}
    _initializer = {"tanh": "glorot_uniform", "sigmoid": "glorot_uniform", "linear": "glorot_uniform",
                    "relu": keras.initializers.he_normal(), "prelu": keras.initializers.he_normal()}
    _optimizer = {"adam": keras.optimizers.Adam, "sgd": keras.optimizers.SGD}
    _regularizer = {"l1": keras.regularizers.l1, "l2": keras.regularizers.l2, "l1_l2": keras.regularizers.l1_l2}
    _requirements = ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad", "momentum", "nesterov", "l1", "l2"]

    """
    Use this class like the following. Note that all keys must match the corresponding tf/keras keys of the layer
    
    ```python
        input_shape = [(65,1,9)]
        output_shape = [(4, )]
        layer_configuration=[
            {"type": "Conv2D", "activation": "relu", "kernel_size": (1, 1), "filters": 8}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "Conv2D", "activation": "relu", "kernel_size": (5, 1), "filters": 16}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "MaxPooling2D", "pool_size": (8, 1), "strides": (1, 1)},
            {"type": "Conv2D", "activation": "relu", "kernel_size": (1, 1), "filters": 16}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "Conv2D", "activation": "relu", "kernel_size": (5, 1), "filters": 32}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "MaxPooling2D", "pool_size": (8, 1), "strides": (1, 1)},
            {"type": "Conv2D", "activation": "relu", "kernel_size": (1, 1), "filters": 32}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "Conv2D", "activation": "relu", "kernel_size": (5, 1), "filters": 64}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "MaxPooling2D", "pool_size": (8, 1), "strides": (1, 1)},
            {"type": "Conv2D", "activation": "relu", "kernel_size": (1, 1), "filters": 64}, 
            {"type": "Dropout", "rate": 0.2},
            {"type": "Flatten"},
            # {"type": "Dense", "units": 128, "activation": "relu"}
        ]
        model = CNNfromConfig(input_shape, output_shape, layer_configuration)
    ```

    """

    def __init__(self, input_shape: list, output_shape: list, layer_configuration: list, optimizer="adam",
                 batch_normalization=False, **kwargs):

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        self.conf = layer_configuration
        activation_output = kwargs.pop("activation_output", "linear")
        self.activation_output = self._activation.get(activation_output)
        self.activation_output_name = activation_output
        self.kwargs = kwargs
        self.bn = batch_normalization
        self.optimizer = self._set_optimizer(optimizer, **kwargs)
        self._layer_save = []

        # apply to model
        self.set_model()
        self.set_compile_options()
        self.set_custom_objects(loss=self.compile_options["loss"][0], var_loss=var_loss)

    def set_model(self):
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = x_input

        for pos, layer_opts in enumerate(self.conf):
            print(layer_opts)
            layer, layer_kwargs, follow_up_layer = self._extract_layer_conf(layer_opts)
            layer_name = self._get_layer_name(layer, layer_kwargs, pos)
            x_in = layer(**layer_kwargs, name=layer_name)(x_in)
            if follow_up_layer is not None:
                for follow_up in to_list(follow_up_layer):
                    layer_name = self._get_layer_name(follow_up, None, pos)
                    x_in = follow_up(name=layer_name)(x_in)
            self._layer_save.append({"layer": layer, **layer_kwargs, "follow_up_layer": follow_up_layer})

        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output(name=f"{self.activation_output_name}_output")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])
        print(self.model.summary())

    @staticmethod
    def _get_layer_name(layer: keras.layers, layer_kwargs: Union[dict, None], pos: int, *args):
        if isinstance(layer, partial):
            name = layer.args[0] if layer.func.__name__ == "Activation" else layer.func.__name__
        else:
            name = layer.__name__
        if "Conv" in name and isinstance(layer_kwargs, dict) and "kernel_size" in layer_kwargs:
            name = name + "_" + "x".join(map(str, layer_kwargs["kernel_size"]))
        if "Pooling" in name and isinstance(layer_kwargs, dict) and "pool_size" in layer_kwargs:
            name = name + "_" + "x".join(map(str, layer_kwargs["pool_size"]))
        name += f"_{pos + 1}"
        return name

    def _set_optimizer(self, optimizer, **kwargs):
        try:
            opt_name = optimizer.lower()
            opt = self._optimizer.get(opt_name)
            opt_kwargs = {}
            if opt_name == "adam":
                opt_kwargs = select_from_dict(kwargs, ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad"])
            elif opt_name == "sgd":
                opt_kwargs = select_from_dict(kwargs, ["lr", "momentum", "decay", "nesterov"])
            return opt(**opt_kwargs)
        except KeyError:
            raise AttributeError(f"Given optimizer {optimizer} is not supported in this model class.")

    def _set_regularizer(self, regularizer, **kwargs):
        if regularizer is None or (isinstance(regularizer, str) and regularizer.lower() == "none"):
            return None
        try:
            reg_name = regularizer.lower()
            reg = self._regularizer.get(reg_name)
            reg_kwargs = {}
            if reg_name in ["l1", "l2"]:
                reg_kwargs = select_from_dict(kwargs, reg_name, remove_none=True)
                if reg_name in reg_kwargs:
                    reg_kwargs["l"] = reg_kwargs.pop(reg_name)
            elif reg_name == "l1_l2":
                reg_kwargs = select_from_dict(kwargs, ["l1", "l2"], remove_none=True)
            return reg(**reg_kwargs)
        except KeyError:
            raise AttributeError(f"Given regularizer {regularizer} is not supported in this model class.")

    def set_compile_options(self):
        # self.compile_options = {"loss": [custom_loss([keras.losses.mean_squared_error, var_loss])],
        #                         "metrics": ["mse", "mae", var_loss]}
        self.compile_options = {"loss": [keras.losses.mean_squared_error],
                                "metrics": ["mse", "mae", var_loss]}

    def _extract_layer_conf(self, layer_opts):
        follow_up_layer = None
        layer_type = layer_opts.pop("type")
        layer = getattr(keras.layers, layer_type, None)
        activation_type = layer_opts.pop("activation", None)
        if activation_type is not None:
            activation = self._activation.get(activation_type)
            kernel_initializer = self._initializer.get(activation_type, "glorot_uniform")
            layer_opts["kernel_initializer"] = kernel_initializer
            follow_up_layer = activation
            if self.bn is True:
                another_layer = keras.layers.BatchNormalization
                if activation_type in ["relu", "linear", "prelu", "leakyrelu"]:
                    follow_up_layer = (another_layer, follow_up_layer)
                else:
                    follow_up_layer = (follow_up_layer, another_layer)
        regularizer_type = layer_opts.pop("kernel_regularizer", None)
        if regularizer_type is not None:
            layer_opts["kernel_regularizer"] = self._set_regularizer(regularizer_type, **self.kwargs)
        return layer, layer_opts, follow_up_layer


class CNN(AbstractModelClass):  # pragma: no cover

    _activation = {"relu": keras.layers.ReLU, "tanh": partial(keras.layers.Activation, "tanh"),
                   "sigmoid": partial(keras.layers.Activation, "sigmoid"),
                   "linear": partial(keras.layers.Activation, "linear"),
                   "selu": partial(keras.layers.Activation, "selu"),
                   "prelu": partial(keras.layers.PReLU, alpha_initializer=keras.initializers.constant(value=0.25)),
                   "leakyrelu": partial(keras.layers.LeakyReLU)}
    _initializer = {"tanh": "glorot_uniform", "sigmoid": "glorot_uniform", "linear": "glorot_uniform",
                    "relu": keras.initializers.he_normal(), "selu": keras.initializers.lecun_normal(),
                    "prelu": keras.initializers.he_normal()}
    _optimizer = {"adam": keras.optimizers.Adam, "sgd": keras.optimizers.SGD}
    _regularizer = {"l1": keras.regularizers.l1, "l2": keras.regularizers.l2, "l1_l2": keras.regularizers.l1_l2}
    _requirements = ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad", "momentum", "nesterov", "l1", "l2"]
    _dropout = {"selu": keras.layers.AlphaDropout}
    _pooling = {"max": keras.layers.MaxPooling2D, "average": keras.layers.AveragePooling2D,
                "mean": keras.layers.AveragePooling2D}

    """
    Define CNN model as in the following examples:
    
    * use same kernel for all layers and use in total 3 conv layers, no dropout or pooling is applied
    
        ```python
        model=CNN,
        kernel_size=5,
        n_layer=3,
        dense_layer_configuration=[128, 64], 
        ```
    
    * specify the kernel sizes, make sure len of kernel size parameter matches number of layers
    
        ```python
        model=CNN,
        kernel_size=[3, 7, 11],
        n_layer=3,
        dense_layer_configuration=[128, 64], 
        ```
        
    * use different number of filters in each layer (can be combined either with fixed or individual kernel sizes), 
      make sure that lengths match. Using layer_configuration always overwrites any value given to n_layers parameter.
        
        ```python
        model=CNN,
        kernel_size=[3, 7, 11],
        layer_configuration=[24, 48, 48],
        ```
    
    * now specify individual kernel sizes and number of filters for each layer
    
        ```python
        model=CNN,
        layer_configuration=[(16, 3), (32, 7), (64, 11)],
        dense_layer_configuration=[128, 64], 
        ```
    
    * add also some dropout and pooling every 2nd layer, dropout is applied after the conv layer, pooling before. Note 
      that pooling will not used in the init layer whereas dropout is already applied there.
      
        ```python
        model=CNN,
        dropout_freq=2,
        dropout=0.3,
        pooling_type="max",
        pooling_freq=2,
        pooling_size=3,
        layer_configuration=[(16, 3), (32, 7), (64, 11)],
        dense_layer_configuration=[128, 64], 
        ```
    """

    def __init__(self, input_shape: list, output_shape: list, activation="relu", activation_output="linear",
                 optimizer="adam", regularizer=None, kernel_size=7, dropout=None, dropout_freq=None, pooling_freq=None,
                 pooling_type="max",
                 n_layer=1, n_filter=10, layer_configuration=None, pooling_size=None,
                 dense_layer_configuration=None, **kwargs):

        assert len(input_shape) == 1
        assert len(output_shape) == 1
        super().__init__(input_shape[0], output_shape[0])

        # settings
        self.activation = self._set_activation(activation)
        self.activation_name = activation
        self.activation_output = self._set_activation(activation_output)
        self.activation_output_name = activation_output
        self.kernel_initializer = self._initializer.get(activation, "glorot_uniform")
        self.kernel_regularizer = self._set_regularizer(regularizer, **kwargs)
        self.kernel_size = kernel_size
        self.optimizer = self._set_optimizer(optimizer, **kwargs)
        self.layer_configuration = (n_layer, n_filter, self.kernel_size) if layer_configuration is None else layer_configuration
        self.dense_layer_configuration = dense_layer_configuration or []
        self.pooling = self._set_pooling(pooling_type)
        self.pooling_size = pooling_size
        self.dropout, self.dropout_rate = self._set_dropout(activation, dropout)
        self.dropout_freq = self._set_layer_freq(dropout_freq)
        self.pooling_freq = self._set_layer_freq(pooling_freq)

        # apply to model
        self.set_model()
        self.set_compile_options()
        # self.set_custom_objects(loss=custom_loss([keras.losses.mean_squared_error, var_loss]), var_loss=var_loss)
        self.set_custom_objects(loss=self.compile_options["loss"][0], var_loss=var_loss)

    def _set_pooling(self, pooling):
        try:
            return self._pooling.get(pooling.lower())
        except KeyError:
            raise AttributeError(f"Given pooling {pooling} is not supported in this model class.")

    def _set_layer_freq(self, param):
        param = 0 if param is None else param
        assert 0 <= param
        assert isinstance(param, int)
        return param

    def _set_activation(self, activation):
        try:
            return self._activation.get(activation.lower())
        except KeyError:
            raise AttributeError(f"Given activation {activation} is not supported in this model class.")

    def _set_optimizer(self, optimizer, **kwargs):
        try:
            opt_name = optimizer.lower()
            opt = self._optimizer.get(opt_name)
            opt_kwargs = {}
            if opt_name == "adam":
                opt_kwargs = select_from_dict(kwargs, ["lr", "beta_1", "beta_2", "epsilon", "decay", "amsgrad"])
            elif opt_name == "sgd":
                opt_kwargs = select_from_dict(kwargs, ["lr", "momentum", "decay", "nesterov"])
            return opt(**opt_kwargs)
        except KeyError:
            raise AttributeError(f"Given optimizer {optimizer} is not supported in this model class.")

    def _set_regularizer(self, regularizer, **kwargs):
        if regularizer is None or (isinstance(regularizer, str) and regularizer.lower() == "none"):
            return None
        try:
            reg_name = regularizer.lower()
            reg = self._regularizer.get(reg_name)
            reg_kwargs = {}
            if reg_name in ["l1", "l2"]:
                reg_kwargs = select_from_dict(kwargs, reg_name, remove_none=True)
                if reg_name in reg_kwargs:
                    reg_kwargs["l"] = reg_kwargs.pop(reg_name)
            elif reg_name == "l1_l2":
                reg_kwargs = select_from_dict(kwargs, ["l1", "l2"], remove_none=True)
            return reg(**reg_kwargs)
        except KeyError:
            raise AttributeError(f"Given regularizer {regularizer} is not supported in this model class.")

    def _set_dropout(self, activation, dropout_rate):
        if dropout_rate is None:
            return None, None
        assert 0 <= dropout_rate < 1
        return self._dropout.get(activation, keras.layers.Dropout), dropout_rate

    def set_model(self):
        """
        Build the model.
        """
        if isinstance(self.layer_configuration, tuple) is True:
            n_layer, n_hidden, kernel_size = self.layer_configuration
            if isinstance(kernel_size, list):
                assert len(kernel_size) == n_layer  # use individual filter sizes for each layer
                conf = [(n_hidden, kernel_size[i]) for i in range(n_layer)]
            else:
                assert isinstance(kernel_size, int)  # use same filter size for all layers
                conf = [(n_hidden, kernel_size) for _ in range(n_layer)]
        else:
            assert isinstance(self.layer_configuration, list) is True
            if not isinstance(self.layer_configuration[0], tuple):
                if isinstance(self.kernel_size, list):
                    assert len(self.kernel_size) == len(self.layer_configuration)  # use individual filter sizes for each layer
                    conf = [(n_filter, self.kernel_size[i]) for i, n_filter in enumerate(self.layer_configuration)]
                else:
                    assert isinstance(self.kernel_size, int)   # use same filter size for all layers
                    conf = [(n_filter, self.kernel_size) for n_filter in self.layer_configuration]
            else:
                assert len(self.layer_configuration[0]) == 2
                conf = self.layer_configuration

        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = x_input
        for layer, (n_filter, kernel_size) in enumerate(conf):
            if self.pooling_size is not None and self.pooling_freq > 0 and layer % self.pooling_freq == 0 and layer > 0:
                x_in = self.pooling((self.pooling_size, 1), strides=(1, 1), padding='valid')(x_in)
            x_in = keras.layers.Conv2D(filters=n_filter, kernel_size=(kernel_size, 1),
                                       kernel_initializer=self.kernel_initializer,
                                       kernel_regularizer=self.kernel_regularizer)(x_in)
            x_in = self.activation()(x_in)
            if self.dropout is not None and self.dropout_freq > 0 and layer % self.dropout_freq == 0:
                x_in = self.dropout(self.dropout_rate)(x_in)

        x_in = keras.layers.Flatten()(x_in)
        for layer, n_hidden in enumerate(self.dense_layer_configuration):
            if n_hidden < self._output_shape:
                break
            x_in = keras.layers.Dense(n_hidden, name=f"Dense_{len(conf) + layer + 1}",
                                      kernel_initializer=self.kernel_initializer, )(x_in)
            x_in = self.activation(name=f"{self.activation_name}_{len(conf) + layer + 1}")(x_in)
            if self.dropout is not None:
                x_in = self.dropout(self.dropout_rate)(x_in)

        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output(name=f"{self.activation_output_name}_output")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])
        print(self.model.summary())

    def set_compile_options(self):
        # self.compile_options = {"loss": [custom_loss([keras.losses.mean_squared_error, var_loss])],
        #                         "metrics": ["mse", "mae", var_loss]}
        self.compile_options = {"loss": [keras.losses.mean_squared_error],
                                "metrics": ["mse", "mae", var_loss]}


class CNN_16_32_64(CNN):

    def set_model(self):
        """
        Build the model.
        """
        x_input = keras.layers.Input(shape=self._input_shape)
        x_in = keras.layers.Conv2D(filters=16, kernel_size=(73, 1),
                                   kernel_initializer=self.kernel_initializer,
                                   kernel_regularizer=self.kernel_regularizer)(x_input)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Conv2D(filters=32, kernel_size=(49, 1),
                                   kernel_initializer=self.kernel_initializer,
                                   kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        if self.dropout is not None:
            x_in = self.dropout(self.dropout_rate)(x_in)
        x_in = keras.layers.MaxPooling2D((25, 1), strides=(1, 1), padding='valid')(x_in)
        x_in = keras.layers.Conv2D(filters=64, kernel_size=(13, 1),
                                   kernel_initializer=self.kernel_initializer,
                                   kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        if self.dropout is not None:
            x_in = self.dropout(self.dropout_rate)(x_in)
        x_in = keras.layers.Flatten()(x_in)
        x_in = keras.layers.Dense(128, kernel_initializer=self.kernel_initializer,
                                  kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(32, kernel_initializer=self.kernel_initializer,
                                  kernel_regularizer=self.kernel_regularizer)(x_in)
        x_in = self.activation()(x_in)
        x_in = keras.layers.Dense(self._output_shape)(x_in)
        out = self.activation_output(name=f"{self.activation_output_name}_output")(x_in)
        self.model = keras.Model(inputs=x_input, outputs=[out])
