__author__ = "Lukas Leufen"
__date__ = "2022-06-27"


from mlair.configuration.path_config import check_path_and_create
from mlair.reference_models.abstract_reference_model import AbstractReferenceModel
import os
import xarray as xr
import pandas as pd


class CAMSforecast(AbstractReferenceModel):

    def __init__(self, ref_name: str, ref_store_path: str = None, data_path: str = None, interp_method: str = None):
        """
        Use parameters `cams_data_path` to set `data_path` and `cams_interp_method` to set `interp_method` in MLAir
        run script.

        :param ref_name:
        :param ref_store_path:
        :param data_path:
        :param interp_method:
        """

        super().__init__()
        self.ref_name = ref_name
        if ref_store_path is None:
            ref_store_path = f"{self.ref_name}/"
        self.ref_store_path = ref_store_path
        if data_path is None:
            self.data_path = os.path.abspath(".")
        else:
            self.data_path = os.path.abspath(data_path)
        self.interp_method = interp_method
        self.file_pattern = "forecasts_%s_test.nc"
        self.time_dim = "index"
        self.ahead_dim = "ahead"
        self.type_dim = "type"

    def make_reference_available_locally(self, stations):
        "dma8eu_ENS_FORECAST_2019-04-09.nc"
        missing_stations = self.list_locally_available_references(self.ref_store_path, stations)
        if len(missing_stations) > 0:
            check_path_and_create(self.ref_store_path)
            dataset = xr.open_mfdataset(os.path.join(self.data_path, "dma8eu_ENS_FORECAST_*.nc"))
            darray = dataset.to_array().sortby(["longitude", "latitude"])
            for station, coords in missing_stations.items():
                lon, lat = coords["lon"], coords["lat"]
                if self.interp_method is None:
                    station_data = darray.sel(longitude=lon, latitude=lat, method="nearest", drop=True).squeeze(drop=True)
                else:
                    station_data = darray.interp(**{"longitude": lon, "latitude": lat}, method=self.interp_method)
                    station_data = station_data.drop_vars(["longitude", "latitude"]).squeeze(drop=True)
                station_data = station_data.expand_dims(dim={self.type_dim: [self.ref_name]}).compute()
                station_data.coords[self.time_dim] = station_data.coords[self.time_dim] - pd.Timedelta(days=1)
                station_data.coords[self.ahead_dim] = station_data.coords[self.ahead_dim] + 1
                file_name = self.file_pattern % str(station)
                station_data.to_netcdf(os.path.join(self.ref_store_path, file_name))

    @staticmethod
    def list_locally_available_references(reference_path, stations) -> dict:
        try:
            file_list = os.listdir(reference_path)
            if len(file_list) > 0:
                res = {k: v for k, v in stations.items() if all(k not in x for x in file_list)}
            else:
                res = stations
        except FileNotFoundError:
            res = stations
        return res
